use std::collections::VecDeque;

/// The type we use to represent values stored in out CPU.
pub type Word = i32;
/// The type we use to adress registers.
pub type Register = usize;

// Operations on register can also support the dereference operator.
// Because of this, every instruction that takes a register has two
// variants: indirect and direct access

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum RegRef {
    Direct(Register),
    Indirect(Register),
}

/// Our whole instruction set. Consult ISA.md for an explaination.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum Instruction {
    READ(RegRef),  // read n
    WRITE(RegRef), // write n
    LOAD(RegRef),  // load =4
    LOADI(Word),   // load 4 (load immediate)
    STORE(RegRef), //
    ADD(RegRef),
    ADDI(Word),
    SUB(RegRef),
    SUBI(Word),
    JUMP(Word),
    JZERO(Word),
    JGTZ(Word),
    JUMPR(RegRef),
    HALT,
}

use Instruction::*;

type InstructionNumber = usize;

#[derive(Debug, Clone)]
pub enum CpuError {
    EndOfInput(InstructionNumber),
    BadRegister(InstructionNumber, Register),
    OverFlow(InstructionNumber),
    AccAccess(InstructionNumber),
    BadJump(InstructionNumber),
    ProgramIsOver(InstructionNumber),
}

use CpuError::*;

pub struct CPU {
    pc: Register,  // Which register is the program counter?
    acc: Register, // Which register is the accumulator?
    program: Vec<Instruction>,
    regmemory: Vec<Word>,
    input_filo: VecDeque<Word>,
    output_filo: VecDeque<Word>,

    acc_access: bool, // can we access the acc register? Defaults to no.
    pub stopped: bool,
    steps: u32, // how many steps did the program take to run?
}

impl Default for CPU {
    fn default() -> Self {
        Self {
            pc: 9,
            acc: 0,
            program: vec![],
            regmemory: vec![0; 10],
            input_filo: vec![].into(),
            output_filo: vec![].into(),
            stopped: false,
            acc_access: false,
            steps: 0,
        }
    }
}

impl CPU {
    /// Exposes the state of the CPU:
    /// Returns a triplet of vectors, corresponding to:
    /// (entire memory, input queue, output queue, current_program)
    pub fn exposestate(&self) -> (Vec<Word>, Vec<Word>, Vec<Word>, Vec<Instruction>) {
        return (
            self.regmemory.clone(),
            self.input_filo.clone().into(),
            self.output_filo.clone().into(),
            self.program.clone(),
        );
    }

    /// Creates a CPU given a program.
    /// Interally, it calls CPU::Default followed by
    /// load_new.
    pub fn load(program: &Vec<Instruction>) -> Self {
        let mut p = Self::default();
        p.load_new(program);
        return p;
    }

    /// Loads a program in memory.
    pub fn load_new(&mut self, program: &Vec<Instruction>) {
        self.program = program.to_vec();
    }

    /// Adds a word to the input queue.
    pub fn add_input(&mut self, w: Word) {
        self.input_filo.push_back(w);
    }

    /// Overwrites the entire input queue.
    pub fn set_input(&mut self, words: Vec<Word>) {
        self.input_filo = words.into();
    }

    /// Used internally, retrives one number from the input queue.
    /// Returns an EndOfInput error if the queue is empty.
    fn getinput(&mut self) -> Result<Word, CpuError> {
        match self.input_filo.pop_back() {
            Some(w) => return Ok(w),
            None => {
                let n = self.readmemory(&self.pc.clone()).unwrap();
                Err(EndOfInput(n as usize))
            }
        }
    }

    /// Used internally, writes one number to the output queue.
    fn writeoutput(&mut self, w: &Word) {
        self.output_filo.push_front(*w)
    }

    /// Increases the program counter by one.
    /// If the program counter ever reaches the end of
    /// the program (no HALT instruction), returns a ProgramIsOver error.
    fn pc_increase(&mut self) -> Result<(), CpuError> {
        self.regmemory[self.pc] += 1;
        if self.regmemory[self.pc] >= self.program.len() as Word {
            let n = self.readmemory(&self.pc.clone()).unwrap();
            return Err(ProgramIsOver(n as usize));
        }
        Ok(())
    }

    /// Check if an address is the accumulator.
    /// Used to check if we're trying to access the accumulator.
    /// Returns an AccAccess error in case. If acc_access is set to true, does nothing.
    fn acc_access_check(&mut self, addr: &Register) -> Result<(), CpuError> {
        if !self.acc_access {
            if *addr == self.acc {
                let n = self.readmemory(&self.pc.clone()).unwrap();
                return Err(AccAccess(n as usize));
            }
        }
        return Ok(());
    }

    /// Check if a jump instruction will take the program to a
    /// non-existing instruction. Returns a BadJump error in case.
    fn jump_check(&mut self, dest: &Word) -> Result<(), CpuError> {
        if *dest > self.program.len() as Word - 1 {
            let n = self.readmemory(&self.pc.clone()).unwrap();
            return Err(BadJump(n as usize));
        } else {
            return Ok(());
        }
    }

    /// Retrives the value of some register. If the register doesn't
    /// exist, returns a BadRegister error.
    fn readmemory(&self, addr: &Register) -> Result<Word, CpuError> {
        if addr >= &self.regmemory.len() {
            let n = self.readmemory(&self.pc.clone()).unwrap();
            return Err(BadRegister(n as usize, *addr));
        }

        return Ok(self.regmemory[*addr]);
    }

    /// Writes a value to a register. If the register doesn't exist, returns a BadRegister.
    fn writememory(&mut self, addr: &Register, w: Word) -> Result<(), CpuError> {
        if addr >= &self.regmemory.len() {
            let n = self.readmemory(&self.pc.clone()).unwrap();
            return Err(BadRegister(n as usize, *addr));
        }

        self.regmemory[*addr] = w;
        return Ok(());
    }

    fn deref(&self, addref: &RegRef) -> Result<Register, CpuError> {
        match addref {
            RegRef::Direct(r) => return Ok(*r),
            RegRef::Indirect(addr) => {
                let reg = self.readmemory(addr)?;
                return Ok(reg as Register);
            }
        }
    }

    /// Executes a single statement.
    pub fn execone(&mut self) -> Result<(), CpuError> {
        let acc = &self.acc.clone(); // constant
        let pc = &self.pc.clone(); // constant

        let pc_val = self.readmemory(pc)? as usize;
        if pc_val + 1 > self.program.len() {
            return Err(ProgramIsOver(pc_val));
        }
        let op = &self.program[pc_val]; // fetch

        // println!("{:?}, {}", *op, pc_val);

        match *op {
            READ(rgref) => {
                let r = self.deref(&rgref)?;
                self.acc_access_check(&r)?;
                let k = self.getinput()?;
                self.writememory(&r, k)?;
            }

            WRITE(rgref) => {
                let r = self.deref(&rgref)?;
                self.acc_access_check(&r)?;
                let k = self.readmemory(&r)?;
                self.writeoutput(&k);
            }

            LOAD(rgref) => {
                let r = self.deref(&rgref)?;
                self.acc_access_check(&r)?;
                let k = self.readmemory(&r)?;
                self.writememory(acc, k)?;
            }

            LOADI(w) => {
                self.writememory(acc, w)?;
            }

            STORE(rgref) => {
                let r = self.deref(&rgref)?;
                self.acc_access_check(&r)?;
                let k = self.readmemory(acc)?;
                self.writememory(&r, k)?;
            }

            ADD(rgref) => {
                let r = self.deref(&rgref)?;
                self.acc_access_check(&r)?;
                let k = &self.readmemory(&r)?;
                let acc_current = self.readmemory(acc)?;
                self.writememory(acc, acc_current + k)?;
            }

            ADDI(w) => {
                let acc_current = self.readmemory(acc)?;
                self.writememory(acc, acc_current + w)?;
            }

            SUB(rgref) => {
                let r = self.deref(&rgref)?;
                self.acc_access_check(&r)?;
                let k = &self.readmemory(&r)?;
                let acc_current = self.readmemory(acc)?;
                self.writememory(acc, acc_current - k)?;
            }

            SUBI(w) => {
                let acc_current = self.readmemory(acc)?;
                self.writememory(acc, acc_current - w)?;
            }

            JUMP(u) => {
                self.jump_check(&u)?;
                self.writememory(pc, u)?;
                return Ok(());
            }

            JZERO(u) => {
                self.jump_check(&u)?;
                let accval = self.readmemory(acc)?;
                if accval == 0 {
                    self.writememory(pc, u)?;
                    return Ok(());
                }
            }

            JGTZ(u) => {
                self.jump_check(&u)?;
                let accval = self.readmemory(acc)?;
                if accval > 0 {
                    self.writememory(pc, u)?;
                    return Ok(());
                }
            }

            // NON DOCUMENTED!
            JUMPR(rgref) => {
                let r = self.deref(&rgref)?;
                let u = self.readmemory(&r)?;
                self.jump_check(&u)?;
                self.writememory(pc, u)?;
                return Ok(());
            }

            HALT => {
                self.stopped = true;
                self.writememory(pc, pc_val as Word + 1)?;
                return Ok(());
            } // manually increase pc, without checking.
        };

        self.pc_increase()?;
        return Ok(());
    }

    /// Executes the whole program, calling execone multiple times
    /// until a HALT instruction is found.
    pub fn exec(&mut self) -> Result<(), CpuError> {
        while !self.stopped {
            self.execone()?;
            self.steps += 1;
        }

        return Ok(());
    }
}
