use crate::parser::{parse_line, parse_program, ParseError};
use js_sys::{Array, Int32Array};
use wasm_bindgen::prelude::*;

use super::cpu::CpuError::*;
use super::cpu::*;

use Instruction::*;
use RegRef::*;

impl std::fmt::Display for RegRef {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        match self {
            RegRef::Direct(k) => write!(f, "{}", k),
            RegRef::Indirect(k) => write!(f, "*{}", k),
        }
    }
}

impl std::fmt::Display for Instruction {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        match self {
            READ(k) => write!(f, "READ {}", k),
            WRITE(k) => write!(f, "WRITE {}", k),
            LOAD(k) => write!(f, "LOAD {}", k),
            LOADI(k) => write!(f, "LOAD ={}", k),
            STORE(k) => write!(f, "STORE {}", k),
            ADD(k) => write!(f, "ADD {}", k),
            ADDI(k) => write!(f, "ADD ={}", k),
            SUB(k) => write!(f, "SUB {}", k),
            SUBI(k) => write!(f, "SUB ={}", k),
            JUMP(k) => write!(f, "JUMP {}", k),
            JZERO(k) => write!(f, "JZERO {}", k),
            JGTZ(k) => write!(f, "JGTZ {}", k),
            JUMPR(k) => write!(f, "JUMPR {}", k),
            HALT => write!(f, "HALT"),
        }
    }
}

#[wasm_bindgen]
extern "C" {
    #[wasm_bindgen(js_namespace = CPUEMU)]
    fn signalerror(msg: &str, l: Option<u32>);
}

#[wasm_bindgen]
extern "C" {

    #[wasm_bindgen(js_namespace = CPUEMU)]
    fn signalparseerr(l: u32);
}

#[wasm_bindgen]
pub struct WrapCpu {
    cpu: CPU, // cpu: Vec<u32>
}

#[wasm_bindgen]
impl WrapCpu {
    #[wasm_bindgen(constructor)]
    pub fn new() -> Self {
        return WrapCpu {
            cpu: CPU::default(),
        };
    }

    pub fn reset(&mut self) {
        self.cpu = CPU::default();
    }

    // view
    pub fn showmem(&self) -> Int32Array {
        let (k, _, _, _) = self.cpu.exposestate();
        return Int32Array::from(&k[..]);
    }

    pub fn showinput(&self) -> Int32Array {
        let (_, k, _, _) = self.cpu.exposestate();
        return Int32Array::from(&k[..]);
    }

    pub fn showoutput(&self) -> Int32Array {
        let (_, _, k, _) = self.cpu.exposestate();
        return Int32Array::from(&k[..]);
    }

    pub fn showprogram(&self) -> Array {
        let (_, _, _, k) = self.cpu.exposestate();
        let arr = Array::new_with_length(k.len() as u32);
        for i in 0..arr.length() {
            let s = JsValue::from_str(&format!("{}: {}", i, k[i as usize]));
            arr.set(i, s);
        }
        return arr;
    }

    pub fn add_input(&mut self, x: i32) {
        self.cpu.add_input(x)
    }

    pub fn set_input(&mut self, x: Int32Array) {
        self.cpu.set_input(x.to_vec());
    }

    pub fn one_step(&mut self) {
        if !self.cpu.stopped {
            match self.cpu.execone() {
                Ok(()) => {}
                Err(k) => cpuErrDispatch(k),
            };
        } else {
            signalerror("The program reached halt. All good!", None);
        }
    }

    pub fn run(&mut self) {
        match self.cpu.exec() {
            Ok(()) => {}
            Err(k) => cpuErrDispatch(k),
        };
    }

    pub fn load_program(&mut self, prg: &str) {
        match parse_program(prg) {
            Ok(p) => self.cpu.load_new(&p),
            Err(ParseError::ErrorAtLine(l)) => signalparseerr(l as u32),
        }
    }
}

fn cpuErrDispatch(err: CpuError) {
    match err {
        EndOfInput(n) => signalerror("End of input! Add more input?", Some(n as u32)),
        BadRegister(n, r) => signalerror(
            &format!("The register {} doesn't exist!", r),
            Some(n as u32),
        ),
        OverFlow(n) => signalerror(
            "Overflow! Did you go into the negative numbers?",
            Some(n as u32),
        ),
        AccAccess(n) => signalerror(
            "Tried to access register 0. That's illegal!",
            Some(n as u32),
        ),
        BadJump(n) => signalerror("Bad jump instruction!", Some(n as u32)),
        ProgramIsOver(n) => signalerror("The program is over! Forgot to HALT?", Some(n as u32)),
    }
}
